"""Test class for base config."""

import sys
import os
import unittest

sys.path.append(os.path.realpath(os.path.dirname(__file__) + '/..'))

from jeedomdaemon.base_config import BaseConfig # pylint: disable=wrong-import-position

class TestBaseConfig(unittest.TestCase):
    def test_base_config_creation(self):
        """
        Test that it can create a basic config parser
        """
        config = BaseConfig()
        config.parse([])
        self.assertEqual(config.socket_host, "127.0.0.1")

    def test_base_config_parse(self):
        """
        Test that it can parse config
        """
        config = BaseConfig()
        config.parse(['--loglevel', 'info', '--socketport', '42000', '--callback', 'http://localhost/path', '--apikey', 'cnysltyql', '--pid', '123'])
        self.assertEqual(config.log_level, "info")
        self.assertEqual(config.socket_host, "127.0.0.1")
        self.assertEqual(config.socket_port, 42000)
        self.assertEqual(config.callback_url, "http://localhost/path")
        self.assertEqual(config.api_key, "cnysltyql")
        self.assertEqual(config.pid_filename, "123")
        self.assertEqual(config.cycle, 0.5)

    def test_custom_config_parse_with_property(self):
        """
        Test that it can parse config
        """
        class TestConfig(BaseConfig):
            def __init__(self):
                super().__init__()
                self.add_argument("--clientId", type=str)

            @property
            def client_id(self):
                return str(self._args.clientId)

        config = TestConfig()
        config.parse(['--clientId', 'hfldhfsd'])
        self.assertEqual(config.client_id, "hfldhfsd")

    def test_custom_config_parse_without_property(self):
        """
        Test that it can parse config
        """
        class TestConfig(BaseConfig):
            def __init__(self):
                super().__init__()
                self.add_argument("--clientId", type=str)

        config = TestConfig()
        config.parse(['--clientId', 'hfldhfsd'])
        self.assertEqual(config.clientId, "hfldhfsd")

if __name__ == '__main__':
    unittest.main()



import logging
import os

class Utils():

    @staticmethod
    def convert_log_level(level='error'):
        LEVELS = {'debug': logging.DEBUG,
                  'info': logging.INFO,
                  'notice': logging.WARNING,
                  'warning': logging.WARNING,
                  'error': logging.ERROR,
                  'critical': logging.CRITICAL,
                  'none': logging.NOTSET}
        return LEVELS.get(level, logging.NOTSET)

    @staticmethod
    def init_logger(level='error'):
        FORMAT = '[%(asctime)-15s][%(levelname)s] : %(message)s'
        logging.basicConfig(level=Utils.convert_log_level(level), format=FORMAT, datefmt="%Y-%m-%d %H:%M:%S")

    @staticmethod
    def write_pid(path):
        pid = str(os.getpid())
        logging.debug("Writing PID %s to %s", pid, path)
        open(path, 'w').write("%s\n" % pid)



"""Test class for base config."""

import sys
import os
import unittest

sys.path.append(os.path.realpath(os.path.dirname(__file__) + '/..'))

from jeedomdaemon.base_daemon import BaseDaemon # pylint: disable=wrong-import-position
from jeedomdaemon.base_config import BaseConfig # pylint: disable=wrong-import-position

class TestBaseConfig(unittest.TestCase):
    def test_base_daemon_creation(self):
        """
        Test that it can create a basic config parser
        """
        config = BaseConfig()
        config.parse([])
        BaseDaemon(config)

if __name__ == '__main__':
    unittest.main()



"""Module providing a base class to parse config of your daemon."""

from __future__ import annotations

import argparse
from typing import Sequence

class BaseConfig():
    """Base config class, if you need custom configuration you can inherit from this class

    This class support following argument, you don't need to declare them:
    * --loglevel
    * --sockethost
    * --socketport
    * --callback
    * --apikey
    * --pid
    * --cycle

    If you need additionals arguments then simply create a child class and them in your constructor, e.g.:
        ```
        class MyOwnConfig(BaseConfig):
            def __init__(self):
                super().__init__()

                self.add_argument("--user", type=str, default='john')
                self.add_argument("--password", type=str)
        ```
    The first arg is the expected argument from Jeedom, then it's interesting to specify the expected type and if a default value if needed
    """
    def __init__(self):
        self._args = None
        self.__parser = argparse.ArgumentParser(description='Daemon for Jeedom plugin')
        self.add_argument("--loglevel", help="Log Level for the daemon", type=str, default='error')
        self.add_argument("--sockethost", help="Socket host", type=str, default='127.0.0.1')
        self.add_argument("--socketport", help="Socket Port", type=int, default=0)
        self.add_argument("--callback", help="Jeedom callback url", type=str)
        self.add_argument("--apikey", help="Plugin API Key", type=str)
        self.add_argument("--pid", help="daemon pid", type=str)
        self.add_argument("--cycle", help="cycle", type=float, default=0.5)

    def add_argument(self, *args, **kwargs):
        """Add a, argurment to parse.

        e.g. from your child class:
            ```
            self.add_argument("--clientId", type=str)
            self.add_argument("--intValue", type=int, default=5)
            ```
        The first arg is the expected argument from Jeedom, then it's interesting to specify the expected type and a default value if needed
        """
        return self.__parser.add_argument(*args, **kwargs)

    def parse(self, args: Sequence[str] | None = None):
        """Actually parse de config, it will be done for you at daemon start."""
        if self._args is None:
            self._args = self.__parser.parse_args(args)

    def __getattr__(self, name):
        return getattr(self._args, name)

    @property
    def callback_url(self):
        """Return the callback url to Jeedom."""
        return str(self._args.callback)

    @property
    def socket_host(self):
        """Return the daemon socket host."""
        return str(self._args.sockethost)

    @property
    def socket_port(self):
        """Return the daemon socket port."""
        return int(self._args.socketport)

    @property
    def log_level(self):
        """Return the log level."""
        return str(self._args.loglevel)

    @property
    def api_key(self):
        """Return the api key."""
        return str(self._args.apikey)

    @property
    def pid_filename(self):
        """Return the pid."""
        return str(self._args.pid)

    @property
    def cycle(self):
        """Return the cycle."""
        return float(self._args.cycle)



"""Module providing a base class for your daemon."""

from __future__ import annotations

import logging
import signal
import os
import sys
import asyncio
import functools
from typing import Callable, Awaitable

from .utils import Utils
from .aio_connector import Publisher, Listener
from .base_config import BaseConfig

class BaseDaemon:
    """Base daemon class for your daemon

    You should inherit your daemon from it and implement the needed functions amongst on_start, on_message and on_stop

        ```
        class MyDaemon(BaseDaemon):
            def __init__(self) -> None:
                # Standard initialisation
                super().__init__(on_start_cb=self.on_start, on_message_cb=self.on_message, on_stop_cb=self.on_stop)

                # Add here any initialisation your daemon would need

            async def on_start(self):
                # if you don't have specific action to do on start, do not create this method
                pass


            async def on_message(self, message: list):
                pass

            async def on_stop(self):
                # if you don't have specific action to do on stop, do not create this method
                pass
        ```

    To send feeback to Jeedom you have 4 possibilities depending your use case.
    If you are in an async method:
        * await self._publisher.send_to_jeedom(payload) will send a single message with the given payload
        * await self._publisher.add_change(key, value) will add the key/value to the payload of the next cycle

    If not:
        * self._publisher.run_send_to_jeedom(payload) will send a single message with the given payload
        * self._publisher.run_add_change(key, value) will add the key/value to the payload of the next cycle
    """
    def __init__(self,
                 config: BaseConfig = BaseConfig(),
                 on_start_cb: Callable[..., Awaitable[None]] | None = None,
                 on_message_cb: Callable[[list], Awaitable[None]] | None = None,
                 on_stop_cb: Callable[..., Awaitable[None]] | None = None,
                 ) -> None:
        self._config = config
        self._config.parse()
        self.__listen_task: asyncio.Task[None] | None = None
        self._loop = None
        self._publisher: Publisher | None = None
        self._logger = logging.getLogger(__name__)
        self.__log_level = Utils.convert_log_level(self._config.log_level)
        self.__on_start_cb = on_start_cb
        self.__on_message_cb = on_message_cb
        self.__on_stop_cb = on_stop_cb

        Utils.init_logger(self._config.log_level)
        logging.getLogger('asyncio').setLevel(logging.WARNING)

    def set_logger_log_level(self, logger_name: str):
        """ Helper function to set the log level to the given logger"""
        logging.getLogger(logger_name).setLevel(self.log_level)

    @property
    def log_level(self):
        """ Return the log level"""
        return self.__log_level

    def run(self):
        """ Run your daemon, this is the function you should call! """
        try:
            self._logger.info('Starting daemon with log level: %s', self._config.log_level)
            Utils.write_pid(str(self._config.pid_filename))

            asyncio.run(self.__run())
        except Exception as e: # pylint: disable=broad-exception-caught
            self._logger.error('Fatal error: %s', e)
        finally:
            self._logger.info("Shutdown")
            try:
                self._logger.debug("Removing PID file %s", self._config.pid_filename)
                os.remove(self._config.pid_filename)
            except: # pylint: disable=bare-except
                pass

            self._logger.debug("Exit 0")
            sys.stdout.flush()
            sys.exit(0)

    async def __run(self):
        if self._config.socket_port < 1024 or self._config.socket_port>65535:
            raise ValueError()

        self._loop = asyncio.get_running_loop()
        self.__listen_task = Listener.create_listen_task(self._config.socket_host, self._config.socket_port, self.__on_socket_message)

        async with Publisher(self._config.callback_url, self._config.api_key) as self._publisher:
            # self._jeedom_publisher = Publisher(self._config.callback_url, self._config.api_key)
            if not await self._publisher.test_callback():
                return

            if self.__on_start_cb is not None and asyncio.iscoroutinefunction(self.__on_start_cb):
                await self.__on_start_cb()

            self._publisher.create_send_task()

            await self.__add_signal_handler()
            await asyncio.sleep(1) # allow  all tasks to start

            await self.__listen_task

    async def stop(self):
        """ Stop your daemon if need be"""

        if self.__on_stop_cb is not None:
            await self.__on_stop_cb()

        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        self._logger.info("Cancelling %i outstanding tasks", len(tasks))
        for task in tasks:
            task.cancel()
        try:
            asyncio.gather(*tasks, return_exceptions=True)
        except BaseException as e: # pylint: disable=broad-exception-caught
            self._logger.warning("Some exception occured during cancellation: %s", e)

    def __ask_exit(self, sig):
        self._logger.info("Signal %i caught, exiting...", sig)
        asyncio.run_coroutine_threadsafe(self.stop(), self._loop)

    async def __add_signal_handler(self):
        self._loop.add_signal_handler(signal.SIGINT, functools.partial(self.__ask_exit, signal.SIGINT))
        self._loop.add_signal_handler(signal.SIGTERM, functools.partial(self.__ask_exit, signal.SIGTERM))

    async def __on_socket_message(self, message):
        if message['apikey'] != self._config.api_key:
            self._logger.error('Invalid apikey from socket : %s', message)
            return
        try:
            if self.__on_message_cb is not None:
                await self.__on_message_cb(message)

        except Exception as e: # pylint: disable=broad-exception-caught
            self._logger.error('Send command to demon error: %s', e)



# pylint: disable=missing-module-docstring

from setuptools import setup, find_packages

__version__ = "0.8.9"

setup(
    # Needed to silence warnings (and to be a worthwhile package)
    name='jeedomdaemon',
    url='https://github.com/Mips2648/jeedom-daemon-py',
    author='Mips',
    # author_email='',
    # Needed to actually package something
    packages=find_packages(),
    # Needed for dependencies
    install_requires=['aiohttp'],
    # *strongly* suggested for sharing
    version=__version__,
    # The license can be anything you like
    license='MIT',
    description='A base to implement Jeedom daemon in python',
    # We will also need a readme eventually (there will be a warning)
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords = ['JEEDOM', 'DAEMON', 'ASYNCIO'],
    classifiers=[
        'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package

        'Intended Audience :: Developers',      # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',   # Again, pick a license

        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.11',
    ],
)



"""Module providing Listener & Publisher class for your daemon."""

import logging
import json
import asyncio
from typing import Callable, Awaitable
from collections.abc import Mapping
import aiohttp


class Listener():
    """
    This class allow to create an asyncio task that will open a socket server and listen to it until task is canceled.
    `on_message` call_back will be call with the message as a list as argument
    """
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super().__new__(cls)
        return cls.instance

    def __init__(self, socket_host: str, socket_port: int, on_message_cb: Callable[[list], Awaitable[None]]) -> None:
        self._socket_host = socket_host
        self._socket_port = socket_port
        self._on_message_cb = on_message_cb
        self._logger = logging.getLogger(__name__)

    @staticmethod
    def create_listen_task(socket_host: str, socket_port: int, on_message_cb: Callable[[list], Awaitable[None]]):
        """ Helper function to create the listen task"""
        listener = Listener(socket_host, socket_port, on_message_cb)
        return asyncio.create_task(listener.listen())

    async def listen(self):
        """ listen function, as task should be made out of it. Don't use this function directly by use `create_listen_task()` instead"""
        try:
            server = await asyncio.start_server(self.__handle_read, self._socket_host, port=self._socket_port)

            async with server:
                self._logger.info('Listening on %s:%s', self._socket_host, self._socket_port)
                await server.serve_forever()
        except asyncio.CancelledError:
            self._logger.info("Listening cancelled")

    async def __handle_read(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        data = await reader.read()
        message = data.decode()
        addr = writer.get_extra_info('peername')
        self._logger.debug("Received %s from %s", message, addr)

        writer.close()
        self._logger.debug("Close connection")
        await writer.wait_closed()
        await self._on_message_cb(json.loads(message))

class Publisher():
    """This class allow to push information to Jeedom either immediately by calling function `send_to_jeedom` or in cycle by calling function `add_change`. For the "cycle" mode, a task must be created by calling `create_send_task` and awaited"""
    def __init__(self, callback_url: str, api_key: str, cycle: float = 0.5) -> None:
        self._jeedom_session = aiohttp.ClientSession()
        self._callback_url = callback_url
        self._api_key = api_key
        self._cycle = cycle if (cycle > 0 and cycle < 10) else 0.5
        self._logger = logging.getLogger(__name__)

        self.__changes = {}

    async def __aenter__(self):
        return self

    async def __aexit__(self, *excinfo):
        await self._jeedom_session.close()

    def create_send_task(self):
        """ Helper function to create the send task"""
        return asyncio.create_task(self.__send_task())

    async def test_callback(self):
        """test_callback will return true if communication with Jeedom is sucessfull or false otherwise"""
        try:
            async with self._jeedom_session.get(self._callback_url + '?test=1&apikey=' + self._api_key) as resp:
                if resp.status != 200:
                    self._logger.error("Please check your network configuration page: %s-%s", resp.status, resp.reason)
                    return False
        except aiohttp.ClientError as e:
            self._logger.error('Callback error: %s. Please check your network configuration page', e)
            return False
        return True

    async def __send_task(self):
        self._logger.info("Send async started")
        try:
            last_send_on_error = False
            while True:
                if len(self.__changes)>0:
                    changes = self.__changes
                    self.__changes = {}

                    try:
                        if not await self.send_to_jeedom(changes):
                            await self.__merge_dict(self.__changes,changes)
                    except aiohttp.ClientError as e:
                        if last_send_on_error:
                            self._logger.error("error during send: %s", e)
                        else:
                            self._logger.debug("first time error during send: %s", e)
                            last_send_on_error = True
                        await self.__merge_dict(self.__changes,changes)
                    else:
                        last_send_on_error = False
                await asyncio.sleep(self._cycle)
        except asyncio.CancelledError:
            self._logger.info("Send async cancelled")

    def run_send_to_jeedom(self, payload):
        """
        Will run coroutine to send the payload provided.
        A running loop must exist
        """
        _loop = asyncio.get_running_loop()
        return asyncio.run_coroutine_threadsafe(self.send_to_jeedom(payload), _loop)

    async def send_to_jeedom(self, payload):
        """
        Will send the payload provided.
        return true or false if successful
        """
        self._logger.debug('Send to jeedom: %s', payload)
        async with self._jeedom_session.post(self._callback_url + '?apikey=' + self._api_key, json=payload) as resp:
            if resp.status != 200:
                self._logger.error('Error on send request to jeedom, return %s-%s', resp.status, resp.reason)
                return False
        return True

    def run_add_change(self, key: str, value):
        """
        Will run coroutine add a key/value pair to the payload of the next cycle.
        A running loop must exist
        """
        _loop = asyncio.get_running_loop()
        return asyncio.run_coroutine_threadsafe(self.add_change(key, value), _loop)

    async def add_change(self, key: str, value):
        """
        Add a key/value pair to the payload of the next cycle, several level can be provided at once by separating key with `::`
        If a key already exists the value will be replaced by the newest
        """
        if key.find('::') != -1:

            changes = value
            for k in reversed(key.split('::')):
                tmp_changes = {}
                tmp_changes[k] = changes
                changes = tmp_changes
            await self.__merge_dict(self.__changes,changes)
        else:
            self.__changes[key] = value

    async def __merge_dict(self, dic1: dict, dic2: dict):
        for key,val2 in dic2.items():
            val1 = dic1.get(key) # returns None if v1 has no value for this key
            if isinstance(val1, Mapping) and isinstance(val2, Mapping):
                await self.__merge_dict(val1, val2)
            elif not bool(val1) or bool(val2) :
                dic1[key] = val2
